'''
Created on Feb 29, 2012

@author: Jakub Chmelar
'''

# import system modules
import re
import sys
import time
import subprocess
import math

# import SOI module
from SOI import cartesian

# debug turned off by default
DEBUG = False

# timeout for SSH connection (in seconds)
SSH_TIMEOUT = 10

# timeout for ping (in seconds)
PING_TIMEOUT = SSH_TIMEOUT + 30

# SSH user name
SSH_LOGIN = "cesnet_vutbr2"

# set up conversion constants
# latitude/longitude, unit is meters/degree
LAT_COEF = 111.330
LON_COEF = 64.152

# speed of internet, unit is km/ms
SPEED = 4/9.0 * 300

# --- classes ---

class Landmark():
    def __init__(self, data = None):
        if(data):
            self.hostname = data['hostname']
            self.ip_addr = data['ip'] 

            try:                
                self.latitude = float(data['latitude'])
                self.longitude = float(data['longitude'])
            except(ValueError):
                print >> sys.stderr, "Wrong data format for landmark '%s'" % self.hostname
                sys.exit(1)
        else:
            self.hostname = None
            self.ip_addr = None
            self.latitude = None
            self.longitude = None
            
    def ToPoint(self):
        x = self.latitude * LAT_COEF
        y = self.longitude * LON_COEF
        
        return cartesian.Point(x, y)

class Target():
    def __init__(self, data = None):
        if(data):
            self.hostname = data['hostname']
            self.ip_addr = data['ip'] 

            self.latitude = None
            self.longitude = None

            self.raw_data = data 
        else:
            self.hostname = None
            self.ip_addr = None
            self.latitude = None
            self.longitude = None
            
            self.raw_data = None

    def ToPoint(self):
        x = self.latitude * LAT_COEF
        y = self.longitude * LON_COEF
        
        return cartesian.Point(x, y)

# --- methods ---

def Ping(target, landmarks):
    # set up variables
    processes = {}
    rtts = {}

    if(DEBUG):
        print " * Spawning ping via SSH at nodes:"

    # set up timer
    start_time = time.time()

    # spawn SSH subprocesses
    for hostname, lm in landmarks.items():
        # set up SSH arguments
        command = ["ssh", "-o", "ConnectTimeout=%d" % SSH_TIMEOUT, "-o", "StrictHostKeyChecking=no", "%s@%s" % (SSH_LOGIN, lm.ip_addr), "ping", "-i0.5", "-c10", "-q", target.ip_addr]
        
        # spawn SSH to landmark as subprocess
        if(DEBUG):
            print "    - %s" % hostname
        
        processes[hostname] = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        
    if(DEBUG):
        print " * Ping results:"

    # process subprocesses as they finish
    while(len(processes)):
        for hostname, process in processes.items():
            # check if process finished already
            retcode = process.poll()

            # possible return codes
            # None = process still running
            # 0    = ping successful
            # 1    = no response from target
            # 2    = can't send packets (DNS/iface/... error)
            # 255  = SSH failed
            
            if(retcode != None):
                # process output according to return code
                if(retcode == 0):
                    # get STDOUT from process
                    stdout = process.stdout.read()
                    
                    # parse minimal RTT from STDOUT
                    for line in stdout.split("\n"):
                        match = re.findall("^rtt min/avg/max/mdev = ([0-9.]+)/", line)
                        
                        if(match):
                            # get RTT as float
                            rtt = float(match[0])
                            
                            # save into dictionary by hostname
                            rtts[hostname] = rtt

                            if(DEBUG):
                                print "    - DONE '%s' [%.2f ms]" % (hostname, rtt)
  
                else:
                    if(DEBUG):
                        # get STDERR from process
                        stderr = process.stderr.read()
                    
                        print >> sys.stderr, "    - FAIL '%s' [return code %d]" % (hostname, retcode)
                        
                        if(stderr):
                            print >> sys.stderr, "     + STDERR:"
                            for line in stderr.split("\n"):
                                if(line):
                                    print >> sys.stderr, "     | %s" % (line)
                    
                # remove process from dictionary
                processes.pop(hostname)

        # check if we're out of time
        if(time.time() > start_time + PING_TIMEOUT):
            if(DEBUG):
                print " * Ping timeout reached!"
                
            # terminate remaining processes
            for hostname, process in processes.items():
                if(DEBUG):
                    print "    - terminating connection to '%s'" % hostname

                # kill process                    
                process.kill()
                
            # return rtts prematurely
            return rtts
                
        # sleep a while
        time.sleep(0.1)
                
    # return RTTs
    return rtts

def FromPoint(point):
    return (point.x / LAT_COEF, point.y / LON_COEF)

def RttToDistance(rtt):
    return (rtt / 2) * SPEED

def RttsToPosition(rtts, landmarks):
    # set up variables
    circles = []
    
    if(DEBUG):
        print " * Circles:"
    
    # for each landmark that returned RTT
    for hostname, rtt in rtts.items():
        # get distance
        distance = RttToDistance(rtt)
        
        # get landmark
        lm = landmarks[hostname]
        
        # create circle around landmark
        p = lm.ToPoint()
        c = cartesian.Circle(p, distance)
        
        # add circle into list
        circles.append(c)
        
        if(DEBUG):
            print "    - landmark '%s' [%.2f, %.2f] - max. distance: %d km" % (hostname, lm.latitude, lm.longitude, distance)
        
    # get intersections from circles
    intersections = cartesian.GlobalIntersection(circles)

    # convert intersections to polygon
    polygon = cartesian.IntersectionsToPolygon(circles, intersections)

    if(polygon):
        # get polygon area
        area = cartesian.PolygonArea(polygon)
        
        # get polygon centroid point
        centroid = cartesian.PolygonCentroid(polygon)
        
        # get latitude and longitude from centroid
        latitude, longitude = FromPoint(centroid)
        
        if(DEBUG):
            print " * Estimated position:"
            print "    - latitude = %.2f, longitude = %.2f" % (latitude, longitude)
            print "    - area ~= %.1e km^2" % area
            
        return (latitude, longitude, area)
    else:
        if(DEBUG):
            print >> sys.stderr, "Can't locate target - no intersections"
            
        return None
    
def MultiPing(targets, landmarks):
    # set up output list
    output = []
    counter = 1
    
    for hostname, target in targets.items():
        if(DEBUG):
            print "------------------------"
            print "[%i/%i] Target: '%s'" % (counter, len(targets), hostname)
    
        # increase counter
        counter += 1
    
        # save start time for duration
        start_time = time.time()
    
        # get RTTs from landmarks        
        rtts = Ping(target, landmarks)
        
        # got anything?
        if(not rtts):
            if(DEBUG):
                print >> sys.stderr, "No RTTs from target '%s'" % hostname
            continue
        else:
            if(DEBUG):
                print " * Measured RTTs: %i/%i [%d %%]" % (len(rtts), len(landmarks), 100.0 * len(rtts) / len(landmarks))
    
        # position from RTTs
        position = RttsToPosition(rtts, landmarks)
    
        # got anything?
        if(position):
            # get latitude/longitude
            latitude, longitude, area = position
    
            # get radius
            radius = math.sqrt(area/math.pi)
        
            # get duration
            duration = time.time() - start_time
            
            # get raw data from target and change them accordingly
            raw_data = target.raw_data
            
            raw_data['latitude'] = "%.5f" % latitude
            raw_data['longitude'] = "%.5f" % longitude
            raw_data['duration'] = "%.5f" % duration
    
            raw_data['radius'] = "%.2f" % radius
            raw_data['comment'] = "%.2f" % area
            
            # append to output
            output.append(raw_data)
            
        else:
            if(DEBUG):
                print "Unable to estimate position!"
                
    return output
